/* ============================================================
 *
 * This file is a part of kipi-plugins project
 * http://www.digikam.org
 *
 * Date        : 2008-09-25
 * Description : a tool to convert RAW file to DNG
 *
 * Copyright (C) 2008-2013 by Gilles Caulier <caulier dot gilles at gmail dot com>
 * Copyright (C) 2010-2011 by Jens Mueller <tschenser at gmx dot de>
 *
 * This program is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software Foundation;
 * either version 2, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * ============================================================ */

/**
    References about DNG:
    DNG SDK tutorial:    http://www.adobeforums.com/webx/.3bc2944e
                         http://www.adobeforums.com/webx/.3c054bde
    DNG review:          http://www.barrypearson.co.uk/articles/dng/index.htm
    DNG intro:           http://www.adobe.com/digitalimag/pdfs/dng_primer.pdf
                         http://www.adobe.com/products/dng/pdfs/DNG_primer_manufacturers.pdf
    DNG Specification:   http://www.adobe.com/products/dng/pdfs/dng_spec_1_2_0_0.pdf
    TIFF/EP Spec.:       http://www.map.tu.chiba-u.ac.jp/IEC/100/TA2/recdoc/N4378.pdf
    DNG SDK reference:   http://www.thomasdideriksen.dk/misc/File%20Formats/dng_sdk_refman.pdf
    DNG SDK tarball:     ftp://ftp.adobe.com/pub/adobe/dng/dng_sdk_1_2.zip
    DNG users forum:     http://www.adobeforums.com/webx/.3bb5f0ec

    Applications using DNG SDK:
    DNG4PS2:             http://dng4ps2.chat.ru/index_en.html
    CORNERFIX:           http://sourceforge.net/projects/cornerfix
    ADOBE DNG CONVERTER: ftp://ftp.adobe.com/pub/adobe/photoshop/cameraraw/win/4.x
    DNGCONVERT:          https://github.com/jmue/dngconvert
    MOVIE2DNG:           https://elphel.svn.sourceforge.net/svnroot/elphel/tools/Movie2DNG
*/

#include "dngwriter_p.h"

// C ansi includes

extern "C"
{
#include <sys/stat.h>
//#include <utime.h>
}

// Qt includes

#include <QtGui\qimage.h>
#include <QString>
#include <QByteArray>
#include <QFile>
#include <QFileInfo>
#include <QIODevice>
#include <QTemporaryFile>
#include <qdebug.h>
// KDE includes

#include <kdebug.h>

// LibKDcraw includes

#include <libkdcraw/version.h>

// Local includes

#include "dngwriterhost.h"
#include "kpmetadata.h"

#define CHUNK 65536

using namespace KIPIPlugins;

namespace DNGIface
{

DNGWriter::DNGWriter()
    : d(new Private)
{
}

DNGWriter::~DNGWriter()
{
    delete d;
}

void DNGWriter::cancel()
{
    d->cancel = true;
}

void DNGWriter::reset()
{
    d->reset();
}

void DNGWriter::setCompressLossLess(bool b)
{
    d->jpegLossLessCompression = b;
}

bool DNGWriter::compressLossLess() const
{
    return d->jpegLossLessCompression;
}

void DNGWriter::setUpdateFileDate(bool b)
{
    d->updateFileDate = b;
}

bool DNGWriter::updateFileDate() const
{
    return d->updateFileDate;
}

void DNGWriter::setBackupOriginalRawFile(bool b)
{
    d->backupOriginalRawFile = b;
}

bool DNGWriter::backupOriginalRawFile() const
{
    return d->backupOriginalRawFile;
}

void DNGWriter::setPreviewMode(int mode)
{
    d->previewMode = mode;
}

int DNGWriter::previewMode() const
{
    return d->previewMode;
}

void DNGWriter::setInputFile(const QString& filePath)
{
    d->inputFile = filePath;
}

void DNGWriter::setOutputFile(const QString& filePath)
{
    d->outputFile = filePath;
}

QString DNGWriter::inputFile() const
{
    return d->inputFile;
}

QString DNGWriter::outputFile() const
{
    return d->outputFile;
}

// FIXME: Wow, this is a LOOOOOOOONG method... We should try to refactor it.
int DNGWriter::convert()
{	
	//time check
	clock_t before;
	double result;
	before = clock();

    d->cancel = false;

	qDebug() << "convert() call " ;
    try
    {
        if (inputFile().isEmpty())
        {
            qDebug() << "DNGWriter: No input file to convert. Aborted..." ;
            return PROCESSFAILED;
        }

        QFileInfo inputInfo(inputFile());
        QString   dngFilePath = outputFile();

        if (dngFilePath.isEmpty())
        {
            dngFilePath = QString(inputInfo.completeBaseName() + QString(".dng"));
        }

        QFileInfo          outputInfo(dngFilePath);
        QByteArray         rawData;
        DcrawInfoContainer identify;
        DcrawInfoContainer identifyMake;

        // -----------------------------------------------------------------------------------------		
        qDebug() << "DNGWriter: Loading RAW data from " << inputInfo.fileName() ;

        KDcraw rawProcessor;		
		qDebug() << "inputFile() " << inputFile();
		
		//extention check (.SRW)

		QString	extentionTemp;
		extentionTemp.resize(3);
		QString	extentionSRW = "SRW";
		int sizeTemp = inputFile().length();
		qDebug() << "length = " << inputFile().length();
		
		extentionTemp[0] = inputFile()[sizeTemp-3];
		extentionTemp[1] = inputFile()[sizeTemp-2];
		extentionTemp[2] = inputFile()[sizeTemp-1];

		qDebug() << "extentionTemp = " << extentionTemp;

		if( extentionTemp != extentionSRW )
		{
			qDebug() << " Not Supported File Extention ";
			return FILENOTSUPPORTED;
		}

        if (!rawProcessor.rawFileIdentify(identifyMake, inputFile()))
        {			
            qDebug() << "DNGWriter: Reading RAW file failed. Aborted..." ;
            return PROCESSFAILED;
        }
		
		// model name check
		if( (identifyMake.model != QString("NX3000")) && (identifyMake.model != QString("XXXX")) )
		{
			qDebug() << "identifyMake.model = " << identifyMake.model;
			return FILENOTSUPPORTED;
		}
		qDebug() << "11111111111111111";
        dng_rect activeArea;
        // TODO: need to get correct default crop size to avoid artifacts at the borders
        int activeWidth  = 0;
        int activeHeight = 0;
        int outputHeight = 0;
        int outputWidth  = 0;

        if ((identifyMake.orientation == 5) || (identifyMake.orientation == 6))
        {
            outputHeight = identifyMake.outputSize.width();
            outputWidth  = identifyMake.outputSize.height();
        }
        else
        {
            outputHeight = identifyMake.outputSize.height();
            outputWidth  = identifyMake.outputSize.width();
        }

		result = (double)(clock() - before) / CLOCKS_PER_SEC;
		qDebug() << " rawFileIdentify time " << result ;
		before = clock();

		qDebug() << "222222222222222222222";

        if (!rawProcessor.extractRAWData(inputFile(), rawData, identify, 0))
        {
            qDebug() << "DNGWriter: Loading RAW data failed. Aborted..." ;
            return PROCESSFAILED;
        }
		
		
		result = (double)(clock() - before) / CLOCKS_PER_SEC;
		qDebug() << "extractRAWData time " << result ;
		before = clock();
		//**********************SRW > 12bit unpack RAW**********************//
		//rawdata < raw stream
		//const char* inputPath = inputFile().toStdString().c_str();		

		///* resize nxmini
		rawData.clear();
		M7Codec *mc = new M7Codec();
		QString qInputPath = inputFile();
		qDebug() << "test";
		QByteArray tempArray = qInputPath.toLocal8Bit();
		const char* inputPath = tempArray.data();
		qDebug() << "inputPath = " << inputPath;
		if( mc->M7MuDecoding(inputPath) )
		{
			char aInput;
			char bInput;
			qDebug() << "nReadlFileSize = " << mc->nRealFileSize;			
			int rawSize = identify.fullSize.width() * identify.fullSize.height();
		
			for(int j = 0; j < mc->nRealFileSize; j++)
			{
				short temp = mc->ExtStrBuf12Bits[j]; //margin
				aInput = temp;
				bInput = temp >> 8;				
				rawData.push_back(aInput);
				rawData.push_back(bInput);
			}	

			delete mc->ExtStrBuf12Bits;
		} else {
			qDebug() << "m7codec failed. Aborted..." ;
            return PROCESSFAILED;
		}			
		
		result = (double)(clock() - before) / CLOCKS_PER_SEC;
		qDebug() << "12bit unpack RAW time " << result ;
		before = clock();

		/**/

		qDebug() << "rawdata size   " << rawData.size();
		//**********************SRW > 12bit unpack RAW**********************//
		if (d->cancel)
			return PROCESSCANCELED;

        qDebug() << "DNGWriter: Raw data loaded:" ;
        qDebug() << "--- Data Size:     " << rawData.size() << " bytes";
        qDebug() << "--- Date:          " << identify.dateTime.toString(Qt::ISODate);
        qDebug() << "--- Make:          " << identify.make;
        qDebug() << "--- Model:         " << identify.model;
        qDebug() << "--- ImageSize:     " << identify.imageSize.width() << "x" << identify.imageSize.height();
        qDebug() << "--- FullSize:      " << identify.fullSize.width() << "x" << identify.fullSize.height();
        qDebug() << "--- OutputSize:    " << identify.outputSize.width() << "x" << identify.outputSize.height();
        qDebug() << "--- Orientation:   " << identify.orientation;
        qDebug() << "--- Top margin:    " << identify.topMargin;
        qDebug() << "--- Left margin:   " << identify.leftMargin;
        qDebug() << "--- Filter:        " << identify.filterPattern;
        qDebug() << "--- Colors:        " << identify.rawColors;
        qDebug() << "--- Black:         " << identify.blackPoint;
        qDebug() << "--- White:         " << identify.whitePoint;
        qDebug() << "--- CAM->XYZ:" ;

        QString matrixVal;

        for(int i=0; i<4; ++i)
        {
            qDebug() << "                   "
                     << QString().sprintf("%03.4f  %03.4f  %03.4f", identify.cameraXYZMatrix[i][0],
                                                                    identify.cameraXYZMatrix[i][1],
                                                                    identify.cameraXYZMatrix[i][2]);
        }

        // Check if CFA layout is supported by DNG SDK.
        Private::DNGBayerPattern bayerPattern = Private::Unknown;
        uint32 filter                         = 0;
        bool fujiRotate90                     = false;

		//GRGB
        // Standard bayer layouts
        if (identify.filterPattern == QString("GRBGGRBGGRBGGRBG"))
        {
            bayerPattern = Private::Standard;
            filter       = 0;
        }
        else if (identify.filterPattern == QString("RGGBRGGBRGGBRGGB"))
        {
            bayerPattern = Private::Standard;
            filter       = 1;
        }
        else if (identify.filterPattern == QString("BGGRBGGRBGGRBGGR"))
        {
            bayerPattern = Private::Standard;
            filter       = 2;
        }
        else if (identify.filterPattern == QString("GBRGGBRGGBRGGBRG"))
        {
            bayerPattern = Private::Standard;
            filter       = 3;
        }
        else if ((identify.filterPattern == QString("RGBGRGBGRGBGRGBG")) && (identifyMake.make == QString("FUJIFILM")))
        {
            // Fuji layouts
            bayerPattern = Private::Fuji;
            fujiRotate90 = false;
            filter       = 0;
        }
        else if ((identify.filterPattern == QString("RBGGBRGGRBGGBRGG")) && (identifyMake.make == QString("FUJIFILM")))
        {
            // Fuji layouts
            bayerPattern = Private::Fuji;
            fujiRotate90 = true;
            filter       = 0;
        }
        else if ((identify.rawColors == 3)                 &&
                 (identify.filterPattern.isEmpty())        &&
                 //(identify.filterPattern == QString("")) &&
                 ((uint32)rawData.size() == identify.outputSize.width() * identify.outputSize.height() * 3 * sizeof(uint16)))
        {
            bayerPattern = Private::LinearRaw;
        }
        else if (identify.rawColors == 4)           // Four color sensors
        {
            bayerPattern = Private::FourColor;

            if (identify.filterPattern.length() != 16)
            {
                qDebug() << "DNGWriter: Bayer mosaic not supported. Aborted..." ;
                return FILENOTSUPPORTED;
            }

            for (int i = 0; i < 16; ++i)
            {
                filter = filter >> 2;

                if (identify.filterPattern[i] == 'G')
                {
                    filter |= 0x00000000;
                }
                else if (identify.filterPattern[i] == 'M')
                {
                    filter |= 0x40000000;
                }
                else if (identify.filterPattern[i] == 'C')
                {
                    filter |= 0x80000000;
                }
                else if (identify.filterPattern[i] == 'Y')
                {
                    filter |= 0xC0000000;
                }
                else
                {
                    qDebug() << "DNGWriter: Bayer mosaic not supported. Aborted..." ;
                    return FILENOTSUPPORTED;
                }
            }
        }
        else
        {
            qDebug() << "DNGWriter: Bayer mosaic not supported. Aborted..." ;
            return FILENOTSUPPORTED;
        }

        if (true == fujiRotate90)
        {
            if (false == fujiRotate(rawData, identify))
            {
                qDebug() << "Can not rotate fuji image. Aborted...";
                return PROCESSFAILED;
            }

            int tmp      = outputWidth;
            outputWidth  = outputHeight;
            outputHeight = tmp;
        }
		        
		activeArea   = dng_rect(18, 64, 3714, 5600); // nx 3000 sensor spec
		activeWidth  = identify.outputSize.width();
        activeHeight = identify.outputSize.height();
		qDebug() << "activeWidth  = " << activeWidth;
		qDebug() << "activeHeight  = " << activeHeight;

        // Check if number of Raw Color components is supported.
        if ((identify.rawColors != 3) && (identify.rawColors != 4))
        {
            qDebug() << "DNGWriter: Number of Raw color components not supported. Aborted..." ;
            return PROCESSFAILED;
        }

        int width  = identify.outputSize.width();
        int height = identify.outputSize.height();

		qDebug() << "width  = " << width;
		qDebug() << "height  = " << height;

/*      // NOTE: code to hack RAW data extraction raw data

        QString   rawdataFilePath(inputInfo.completeBaseName() + QString(".dat"));
        QFileInfo rawdataInfo(rawdataFilePath);

        QFile rawdataFile(rawdataFilePath);
        if (!rawdataFile.open(QIODevice::WriteOnly))
        {
            qDebug() << "DNGWriter: Cannot open file to write RAW data. Aborted..." ;
            return PROCESSFAILED;
        }
        QDataStream rawdataStream(&rawdataFile);
        rawdataStream.writeRawData(rawData.data(), rawData.size());
        rawdataFile.close();
/**/

        // -----------------------------------------------------------------------------------------

        if (d->cancel)
            return PROCESSCANCELED;

        qDebug() << "DNGWriter: DNG memory allocation and initialization" ;

        dng_memory_allocator memalloc(gDefaultDNGMemoryAllocator);

        dng_rect rect(height, width);
        //DNGWriterHost host(d, &memalloc);
		dng_threaded_host host;

        host.SetSaveDNGVersion(dngVersion_SaveDefault);
        host.SetSaveLinearDNG(false);
        host.SetKeepOriginalFile(true);

        AutoPtr<dng_image> image(new dng_simple_image(rect, (bayerPattern == Private::LinearRaw) ? 3 : 1, ttShort, memalloc));

        if (d->cancel)
            return PROCESSCANCELED;

        // -----------------------------------------------------------------------------------------

        dng_pixel_buffer buffer;

        buffer.fArea       = rect;
        buffer.fPlane      = 0;
        buffer.fPlanes     = bayerPattern == Private::LinearRaw ? 3 : 1;
        buffer.fRowStep    = buffer.fPlanes * width;
        buffer.fColStep    = buffer.fPlanes;
        buffer.fPlaneStep  = 1;
        buffer.fPixelType  = ttShort;
        buffer.fPixelSize  = TagTypeSize(ttShort);
        buffer.fData       = rawData.data();
        image->Put(buffer);

        if (d->cancel)
            return PROCESSCANCELED;

        // -----------------------------------------------------------------------------------------

        qDebug() << "DNGWriter: DNG Negative structure creation" ;

        AutoPtr<dng_negative> negative(host.Make_dng_negative());

        negative->SetDefaultScale(dng_urational(outputWidth, activeWidth), dng_urational(outputHeight, activeHeight)); // 1 , 1	

		//negative->SetDefaultCropOrigin(12, 12); // nx mini 12
		negative->SetDefaultCropOrigin(32, 24); // nx 3000

        negative->SetDefaultCropSize(5472, 3648); //cropsize nx3000 : 5472 3648
        negative->SetActiveArea(activeArea);
		
		negative->SetModelName(identify.model.toLatin1());
        negative->SetLocalName(QString("%1 %2").arg(identify.make, identify.model).toLatin1());
        negative->SetOriginalRawFileName(inputInfo.fileName().toLatin1());
        negative->SetColorChannels(identify.rawColors);

        ColorKeyCode colorCodes[4] = { colorKeyMaxEnum, colorKeyMaxEnum, colorKeyMaxEnum, colorKeyMaxEnum };

        for(int i = 0; i < qMax(4, identify.colorKeys.length()); ++i)
        {
            if (identify.colorKeys[i] == 'R')
            {
                colorCodes[i] = colorKeyRed;
            }
            else if (identify.colorKeys[i] == 'G')
            {
                colorCodes[i] = colorKeyGreen;
            }
            else if (identify.colorKeys[i] == 'B')
            {
                colorCodes[i] = colorKeyBlue;
            }
            else if (identify.colorKeys[i] == 'C')
            {
                colorCodes[i] = colorKeyCyan;
            }
            else if (identify.colorKeys[i] == 'M')
            {
                colorCodes[i] = colorKeyMagenta;
            }
            else if (identify.colorKeys[i] == 'Y')
            {
                colorCodes[i] = colorKeyYellow;
            }
        }

        negative->SetColorKeys(colorCodes[0], colorCodes[1], colorCodes[2], colorCodes[3]);

        switch (bayerPattern)
        {
            case Private::Standard:
                // Standard bayer mosaicing. All work fine there.
                // Bayer CCD mask: http://en.wikipedia.org/wiki/Bayer_filter
                negative->SetBayerMosaic(filter);
                break;
            case Private::Fuji:
                // TODO: Fuji is special case. Need to setup different bayer rules here.
                // It do not work in all settings. Need indeep investiguations.
                // Fuji superCCD: http://en.wikipedia.org/wiki/Super_CCD
                negative->SetFujiMosaic(filter);
                break;
            case Private::FourColor:
                negative->SetQuadMosaic(filter);
                break;
            default:
                break;
        }

        negative->SetWhiteLevel(identify.whitePoint, 0);
        negative->SetWhiteLevel(identify.whitePoint, 1);
        negative->SetWhiteLevel(identify.whitePoint, 2);
        negative->SetWhiteLevel(identify.whitePoint, 3);

        const dng_mosaic_info* const mosaicinfo = negative->GetMosaicInfo();

        if ((mosaicinfo != NULL) && (mosaicinfo->fCFAPatternSize == dng_point(2, 2)))
        {
            negative->SetQuadBlacks(identify.blackPoint + identify.blackPointCh[0],
                                    identify.blackPoint + identify.blackPointCh[1],
                                    identify.blackPoint + identify.blackPointCh[2],
                                    identify.blackPoint + identify.blackPointCh[3]);
        }
        else
        {
            negative->SetBlackLevel(identify.blackPoint, 0);
        }

        negative->SetBaselineExposure(0.0);
        negative->SetBaselineNoise(1.0);
        negative->SetBaselineSharpness(1.0);

        dng_orientation orientation;

        switch (identify.orientation)
        {
            case DcrawInfoContainer::ORIENTATION_180:
                orientation = dng_orientation::Rotate180();
                break;

            case DcrawInfoContainer::ORIENTATION_Mirror90CCW:
                orientation = dng_orientation::Mirror90CCW();
                break;

            case DcrawInfoContainer::ORIENTATION_90CCW:
                orientation = dng_orientation::Rotate90CCW();
                break;

            case DcrawInfoContainer::ORIENTATION_90CW:
                orientation = dng_orientation::Rotate90CW();
                break;

            default:   // ORIENTATION_NONE
                orientation = dng_orientation::Normal();
                break;
        }

        negative->SetBaseOrientation(orientation);
        negative->SetAntiAliasStrength(dng_urational(100, 100));
        negative->SetLinearResponseLimit(1.0);
        negative->SetShadowScale( dng_urational(1, 1) );
        negative->SetAnalogBalance(dng_vector_3(1.0, 1.0, 1.0));

        // -------------------------------------------------------------------------------

        AutoPtr<dng_camera_profile> prof(new dng_camera_profile);
        prof->SetName(QString("%1 %2").arg(identify.make, identify.model).toLatin1());

        // Set Camera->XYZ Color matrix as profile.
        dng_matrix matrix;
		dng_matrix matrix1;


		//color matrix nx 3000
		dng_matrix_3by3 camXYZ;
		camXYZ[0][0] = 0.7696;
		camXYZ[0][1] = -0.3038;
		camXYZ[0][2] = -0.0601;
		camXYZ[1][0] = -0.4552;
		camXYZ[1][1] = 1.3752 ;
		camXYZ[1][2] = 0.0782;
		camXYZ[2][0] = -0.1007;
		camXYZ[2][1] = 0.2486 ;
		camXYZ[2][2] = 0.4288;
		matrix = camXYZ;

		//Forward matrix nx 3000
		dng_matrix_3by3 forwardMatrixS;
		forwardMatrixS[0][0] = 0.7144;
		forwardMatrixS[0][1] = 0.2095;
		forwardMatrixS[0][2] = 0.0404;
		forwardMatrixS[1][0] = 0.3212;
		forwardMatrixS[1][1] = 0.8383;
		forwardMatrixS[1][2] = -0.1594;
		forwardMatrixS[2][0] = 0.0707;
		forwardMatrixS[2][1] = -0.1480;
		forwardMatrixS[2][2] = 0.9024;
		matrix1 = forwardMatrixS;


		dng_tone_curve tonCurveS;
		tonCurveS.fCoord.resize(7);
		//std::vector<dng_point_real64> inputTonCurve;
		//inputTonCurve.resize(7);		

		dng_point_real64 point;		
		point.v=0.000000000000000000; point.h=0.000000000000000000;		
		tonCurveS.fCoord[0] = point;
		point.v=0.019607843831181526; point.h=0.019607843831181526;
		tonCurveS.fCoord[1] = point;
		point.v=0.098039217293262482; point.h=0.058823529630899429;
		tonCurveS.fCoord[2] = point;
		point.v=0.41960784792900085; point.h=0.19607843458652496;
		tonCurveS.fCoord[3] = point;
		point.v=0.70980393886566162; point.h=0.39215686917304993;
		tonCurveS.fCoord[4] = point;
		point.v=0.95294117927551270; point.h=0.78431373834609985;
		tonCurveS.fCoord[5] = point;
		point.v=1.0000000000000000; point.h=1.0000000000000000;
		tonCurveS.fCoord[6] = point;
		//h
		//v 

		prof->SetColorMatrix1((dng_matrix) matrix);
		prof->SetCalibrationIlluminant1(lsD65);
		prof->SetForwardMatrix1((dng_matrix) matrix1);
		prof->SetToneCurve(tonCurveS);
        negative->AddProfile(prof);

        dng_vector camNeutral(identify.rawColors);

        for (int i = 0; i < identify.rawColors; ++i)
        {
            camNeutral[i] = 1.0/identify.cameraMult[i];
        }

        negative->SetCameraNeutral(camNeutral);

        if (d->cancel)
            return PROCESSCANCELED;

        // -----------------------------------------------------------------------------------------

        qDebug() << "DNGWriter: Updating metadata to DNG Negative" ;
		

        dng_exif* const exif = negative->GetExif();
        exif->fModel.Set_ASCII(identify.model.toLatin1());
        exif->fMake.Set_ASCII(identify.make.toLatin1());

        QString    str;
        KPMetadata meta;		
		
        if (meta.load(inputFile()))
        {
            long int   num, den;
            long       val;

            // Time from original shot
            dng_date_time_info dti;
            dti.SetDateTime(d->dngDateTime(meta.getImageDateTime()));
            exif->fDateTimeOriginal = dti;

            dti.SetDateTime(d->dngDateTime(meta.getDigitizationDateTime(true)));
            exif->fDateTimeDigitized = dti;

            negative->UpdateDateTime(dti);

            // String Tags
            str = meta.getExifTagString("Exif.Image.Make");
            if (!str.isEmpty()) exif->fMake.Set_ASCII(str.trimmed().toLatin1());

            str = meta.getExifTagString("Exif.Image.Model");
            if (!str.isEmpty()) exif->fModel.Set_ASCII(str.trimmed().toLatin1());

            str = meta.getExifTagString("Exif.Image.Software");
            if (!str.isEmpty()) exif->fSoftware.Set_ASCII(str.trimmed().toLatin1());

            str = meta.getExifTagString("Exif.Image.ImageDescription");
            if (!str.isEmpty()) exif->fImageDescription.Set_ASCII(str.trimmed().toLatin1());

            str = meta.getExifTagString("Exif.Image.Artist");
            if (!str.isEmpty()) exif->fArtist.Set_ASCII(str.trimmed().toLatin1());

            str = meta.getExifTagString("Exif.Image.Copyright");
            if (!str.isEmpty()) exif->fCopyright.Set_ASCII(str.trimmed().toLatin1());

            str = meta.getExifTagString("Exif.Photo.UserComment");
            if (!str.isEmpty()) exif->fUserComment.Set_ASCII(str.trimmed().toLatin1());

            str = meta.getExifTagString("Exif.Image.CameraSerialNumber");
            if (!str.isEmpty()) exif->fCameraSerialNumber.Set_ASCII(str.trimmed().toLatin1());

            str = meta.getExifTagString("Exif.GPSInfo.GPSLatitudeRef");
            if (!str.isEmpty()) exif->fGPSLatitudeRef.Set_ASCII(str.trimmed().toLatin1());

            str = meta.getExifTagString("Exif.GPSInfo.GPSLongitudeRef");
            if (!str.isEmpty()) exif->fGPSLongitudeRef.Set_ASCII(str.trimmed().toLatin1());

            str = meta.getExifTagString("Exif.GPSInfo.GPSSatellites");
            if (!str.isEmpty()) exif->fGPSSatellites.Set_ASCII(str.trimmed().toLatin1());

            str = meta.getExifTagString("Exif.GPSInfo.GPSStatus");
            if (!str.isEmpty()) exif->fGPSStatus.Set_ASCII(str.trimmed().toLatin1());

            str = meta.getExifTagString("Exif.GPSInfo.GPSMeasureMode");
            if (!str.isEmpty()) exif->fGPSMeasureMode.Set_ASCII(str.trimmed().toLatin1());

            str = meta.getExifTagString("Exif.GPSInfo.GPSSpeedRef");
            if (!str.isEmpty()) exif->fGPSSpeedRef.Set_ASCII(str.trimmed().toLatin1());

            str = meta.getExifTagString("Exif.GPSInfo.GPSTrackRef");
            if (!str.isEmpty()) exif->fGPSTrackRef.Set_ASCII(str.trimmed().toLatin1());

            str = meta.getExifTagString("Exif.GPSInfo.GPSSpeedRef");
            if (!str.isEmpty()) exif->fGPSSpeedRef.Set_ASCII(str.trimmed().toLatin1());

            str = meta.getExifTagString("Exif.GPSInfo.GPSImgDirectionRef");
            if (!str.isEmpty()) exif->fGPSSpeedRef.Set_ASCII(str.trimmed().toLatin1());

            str = meta.getExifTagString("Exif.GPSInfo.GPSMapDatum");
            if (!str.isEmpty()) exif->fGPSMapDatum.Set_ASCII(str.trimmed().toLatin1());

            str = meta.getExifTagString("Exif.GPSInfo.GPSDestLatitudeRef");
            if (!str.isEmpty()) exif->fGPSDestLatitudeRef.Set_ASCII(str.trimmed().toLatin1());

            str = meta.getExifTagString("Exif.GPSInfo.GPSDestLongitudeRef");
            if (!str.isEmpty()) exif->fGPSDestLongitudeRef.Set_ASCII(str.trimmed().toLatin1());

            str = meta.getExifTagString("Exif.GPSInfo.GPSDestBearingRef");
            if (!str.isEmpty()) exif->fGPSDestBearingRef.Set_ASCII(str.trimmed().toLatin1());

            str = meta.getExifTagString("Exif.GPSInfo.GPSDestDistanceRef");
            if (!str.isEmpty()) exif->fGPSDestDistanceRef.Set_ASCII(str.trimmed().toLatin1());

            str = meta.getExifTagString("Exif.GPSInfo.GPSProcessingMethod");
            if (!str.isEmpty()) exif->fGPSProcessingMethod.Set_ASCII(str.trimmed().toLatin1());

            str = meta.getExifTagString("Exif.GPSInfo.GPSAreaInformation");
            if (!str.isEmpty()) exif->fGPSAreaInformation.Set_ASCII(str.trimmed().toLatin1());

            str = meta.getExifTagString("Exif.GPSInfo.GPSDateStamp");
            if (!str.isEmpty()) exif->fGPSDateStamp.Set_ASCII(str.trimmed().toLatin1());

            // Rational Tags

            if (meta.getExifTagRational("Exif.Photo.ExposureTime", num, den))          exif->fExposureTime             = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.Photo.FNumber", num, den))               exif->fFNumber                  = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.Photo.ShutterSpeedValue", num, den))     exif->fShutterSpeedValue        = dng_srational(num, den);
            if (meta.getExifTagRational("Exif.Photo.ApertureValue", num, den))         exif->fApertureValue            = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.Photo.BrightnessValue", num, den))       exif->fBrightnessValue          = dng_srational(num, den);
            if (meta.getExifTagRational("Exif.Photo.ExposureBiasValue", num, den))     exif->fExposureBiasValue        = dng_srational(num, den);
            if (meta.getExifTagRational("Exif.Photo.MaxApertureValue", num, den))      exif->fMaxApertureValue         = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.Photo.FocalLength", num, den))           exif->fFocalLength              = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.Photo.DigitalZoomRatio", num, den))      exif->fDigitalZoomRatio         = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.Photo.SubjectDistance", num, den))       exif->fSubjectDistance          = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.Image.BatteryLevel", num, den))          exif->fBatteryLevelR            = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.Photo.FocalPlaneXResolution", num, den)) exif->fFocalPlaneXResolution    = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.Photo.FocalPlaneYResolution", num, den)) exif->fFocalPlaneYResolution    = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.GPSInfo.GPSAltitude", num, den))         exif->fGPSAltitude              = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.GPSInfo.GPSDOP", num, den))              exif->fGPSDOP                   = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.GPSInfo.GPSSpeed", num, den))            exif->fGPSSpeed                 = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.GPSInfo.GPSTrack", num, den))            exif->fGPSTrack                 = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.GPSInfo.GPSImgDirection", num, den))     exif->fGPSImgDirection          = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.GPSInfo.GPSDestBearing", num, den))      exif->fGPSDestBearing           = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.GPSInfo.GPSDestDistance", num, den))     exif->fGPSDestDistance          = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.GPSInfo.GPSLatitude", num, den, 0))      exif->fGPSLatitude[0]           = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.GPSInfo.GPSLatitude", num, den, 1))      exif->fGPSLatitude[1]           = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.GPSInfo.GPSLatitude", num, den, 2))      exif->fGPSLatitude[2]           = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.GPSInfo.GPSLongitude", num, den, 0))     exif->fGPSLongitude[0]          = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.GPSInfo.GPSLongitude", num, den, 1))     exif->fGPSLongitude[1]          = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.GPSInfo.GPSLongitude", num, den, 2))     exif->fGPSLongitude[2]          = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.GPSInfo.GPSTimeStamp", num, den, 0))     exif->fGPSTimeStamp[0]          = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.GPSInfo.GPSTimeStamp", num, den, 1))     exif->fGPSTimeStamp[1]          = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.GPSInfo.GPSTimeStamp", num, den, 2))     exif->fGPSTimeStamp[2]          = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.GPSInfo.GPSDestLatitude", num, den, 0))  exif->fGPSDestLatitude[0]       = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.GPSInfo.GPSDestLatitude", num, den, 1))  exif->fGPSDestLatitude[1]       = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.GPSInfo.GPSDestLatitude", num, den, 2))  exif->fGPSDestLatitude[2]       = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.GPSInfo.GPSDestLongitude", num, den, 0)) exif->fGPSDestLongitude[0]      = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.GPSInfo.GPSDestLongitude", num, den, 1)) exif->fGPSDestLongitude[1]      = dng_urational(num, den);
            if (meta.getExifTagRational("Exif.GPSInfo.GPSDestLongitude", num, den, 2)) exif->fGPSDestLongitude[2]      = dng_urational(num, den);

            // Integer Tags

            if (meta.getExifTagLong("Exif.Photo.ExposureProgram", val))                exif->fExposureProgram          = (uint32)val;
            if (meta.getExifTagLong("Exif.Photo.ISOSpeedRatings", val))                exif->fISOSpeedRatings[0]       = (uint32)val;
            if (meta.getExifTagLong("Exif.Photo.MeteringMode", val))                   exif->fMeteringMode             = (uint32)val;
            if (meta.getExifTagLong("Exif.Photo.LightSource", val))                    exif->fLightSource              = (uint32)val;
            if (meta.getExifTagLong("Exif.Photo.Flash", val))                          exif->fFlash                    = (uint32)val;
            if (meta.getExifTagLong("Exif.Photo.SensingMethod", val))                  exif->fSensingMethod            = (uint32)val;
            if (meta.getExifTagLong("Exif.Photo.FileSource", val))                     exif->fFileSource               = (uint32)val;
            if (meta.getExifTagLong("Exif.Photo.SceneType", val))                      exif->fSceneType                = (uint32)val;
            if (meta.getExifTagLong("Exif.Photo.CustomRendered", val))                 exif->fCustomRendered           = (uint32)val;
            if (meta.getExifTagLong("Exif.Photo.ExposureMode", val))                   exif->fExposureMode             = (uint32)val;
            if (meta.getExifTagLong("Exif.Photo.WhiteBalance", val))                   exif->fWhiteBalance             = (uint32)val;
            if (meta.getExifTagLong("Exif.Photo.SceneCaptureType", val))               exif->fSceneCaptureType         = (uint32)val;
            if (meta.getExifTagLong("Exif.Photo.GainControl", val))                    exif->fGainControl              = (uint32)val;
            if (meta.getExifTagLong("Exif.Photo.Contrast", val))                       exif->fContrast                 = (uint32)val;
            if (meta.getExifTagLong("Exif.Photo.Saturation", val))                     exif->fSaturation               = (uint32)val;
            if (meta.getExifTagLong("Exif.Photo.Sharpness", val))                      exif->fSharpness                = (uint32)val;
            if (meta.getExifTagLong("Exif.Photo.SubjectDistanceRange", val))           exif->fSubjectDistanceRange     = (uint32)val;
            if (meta.getExifTagLong("Exif.Photo.FocalLengthIn35mmFilm", val))          exif->fFocalLengthIn35mmFilm    = (uint32)val;
            if (meta.getExifTagLong("Exif.Photo.ComponentsConfiguration", val))        exif->fComponentsConfiguration  = (uint32)val;
            if (meta.getExifTagLong("Exif.Photo.PixelXDimension", val))                exif->fPixelXDimension          = (uint32)val;
            if (meta.getExifTagLong("Exif.Photo.PixelYDimension", val))                exif->fPixelYDimension          = (uint32)val;
            if (meta.getExifTagLong("Exif.Photo.FocalPlaneResolutionUnit", val))       exif->fFocalPlaneResolutionUnit = (uint32)val;
            if (meta.getExifTagLong("Exif.GPSInfo.GPSAltitudeRef", val))               exif->fGPSAltitudeRef           = (uint32)val;
            if (meta.getExifTagLong("Exif.GPSInfo.GPSDifferential", val))              exif->fGPSDifferential          = (uint32)val;
            long gpsVer1 = 0;
            long gpsVer2 = 0;
            long gpsVer3 = 0;
            long gpsVer4 = 0;
            meta.getExifTagLong("Exif.GPSInfo.GPSVersionID", gpsVer1, 0);
            meta.getExifTagLong("Exif.GPSInfo.GPSVersionID", gpsVer2, 1);
            meta.getExifTagLong("Exif.GPSInfo.GPSVersionID", gpsVer3, 2);
            meta.getExifTagLong("Exif.GPSInfo.GPSVersionID", gpsVer4, 3);
			long gpsVer  = 0;
			gpsVer += gpsVer1 << 24;
			gpsVer += gpsVer2 << 16;
			gpsVer += gpsVer3 <<  8;
			gpsVer += gpsVer4;
			exif->fGPSVersionID = (uint32)gpsVer;

			qDebug() <<"LensID = " <<  meta.getExifTagString("Exif.Samsung2.LensType");			
			qDebug() <<"LensFirmware = " <<  meta.getExifTagString("Exif.Samsung2.LensFirmware");
			qDebug() <<"LensSerialNumber = " <<  meta.getExifTagString("Exif.Samsung2.0xa005");

			str = meta.getExifTagString("Exif.Samsung2.LensType");
			exif->fLensID.Set_ASCII(str.trimmed().toLatin1());

			str = meta.getExifTagString("Exif.Photo.LensModel");
			exif->fLensName.Set_ASCII(str.trimmed().toLatin1());

			qDebug() << "fLensID   " <<exif->fLensID.Get() ;
			qDebug() << "fLensName   " <<exif->fLensName.Get() ;
			qDebug() << "fLensSerialNumber   " <<exif->fLensSerialNumber.Get() ;

			//********************************************LDC********************************************//	

			long SRWEncryptionKey[11] = {305, 72, 737, 456, 282, 307, 519, 724, 13, 505, 193};
			long SRWEncryptionKey1[8] = {305, 72, 737, 456, 282, 307, 519, 724}; //distortion key +
			long SRWEncryptionKey2[7] = {724, 13, 505, 193, 305, 72, 737}; //RED key -
			long SRWEncryptionKey3[7] = {456, 282, 307, 519, 724, 13, 13}; //BLUE key -
			long distortion_meta[8];			
			long ColorAbeR_meta[7];
			long ColorAbeB_meta[7];
			double distortion[8];			
			double ColorAbeR[7];
			double ColorAbeB[7];
			int ColorAbeR_Flag = true;
			int ColorAbeB_Flag = true;
			for(int i = 0; i < 8; i++)			
			{
				meta.getExifTagLong("Exif.Samsung2.0xa050", distortion_meta[i] , i);
				distortion[i] = distortion_meta[i] + SRWEncryptionKey1[i];
			}
			for(int i = 0; i < 7; i++)
			{
				meta.getExifTagLong("Exif.Samsung2.0xa051", ColorAbeR_meta[i] , 8+i);
				ColorAbeR[i] = ColorAbeR_meta[i] - SRWEncryptionKey2[i];
			}
			for(int i = 0; i < 7; i++)
			{
				meta.getExifTagLong("Exif.Samsung2.0xa051", ColorAbeB_meta[i] , 15+i);
				ColorAbeB[i] = ColorAbeB_meta[i] - SRWEncryptionKey3[i];
			}
			
			if( ColorAbeR[0] > 10000 )
			{
				ColorAbeR[0] = ColorAbeR[0] - 10000;
				ColorAbeR[0] = ColorAbeR[0] / 20000;
				ColorAbeR_Flag = true;
			}
			else
			{
				ColorAbeR[0] = 10000 - ColorAbeR[0];
				ColorAbeR[0] = ColorAbeR[0] / 20000;
				ColorAbeR_Flag = false;
			}

			if( ColorAbeB[0] > 10000 )
			{
				ColorAbeB[0] = ColorAbeB[0] - 10000;
				ColorAbeB[0] = ColorAbeB[0] / 20000;
				ColorAbeB_Flag = true;
			}
			else
			{
				ColorAbeB[0] = 10000 - ColorAbeB[0];
				ColorAbeB[0] = ColorAbeB[0] / 20000;
				ColorAbeB_Flag = false;
			}
		

			double Gdata[4];
			Gdata[0] = distortion[0]*pow(10,distortion[1]);
			Gdata[1] = distortion[2]*pow(10,distortion[3]); 
			Gdata[2] = distortion[4]*pow(10,distortion[5]);		


			Gdata[0] = Gdata[0];
			Gdata[1] = Gdata[1] * pow(5900.0  , 2); 
			Gdata[2] = Gdata[2] * pow(5900.0  , 4); 
			Gdata[3] = 0.000000;
			
			double Rdata[4];
			if( ColorAbeR_Flag ) // ColorAbeR > 10000 
			{
				Rdata[0] = Gdata[0] + (Gdata[0] * ColorAbeR[0]);
				Rdata[1] = Gdata[1] + (Gdata[1] * ColorAbeR[0]);
				Rdata[2] = Gdata[2] + (Gdata[2] * ColorAbeR[0]);
			}
			else // ColorAbeR < 10000 
			{
				Rdata[0] = Gdata[0] - (Gdata[0] * ColorAbeR[0]);
				Rdata[1] = Gdata[1] - (Gdata[1] * ColorAbeR[0]);
				Rdata[2] = Gdata[2] - (Gdata[2] * ColorAbeR[0]);
			}
			Rdata[3] = 0.000000;

			double Bdata[4];
			if( ColorAbeB_Flag ) // ColorAbeB_Flag > 10000 
			{
				Bdata[0] = Gdata[0] + (Gdata[0] * ColorAbeB[0]);
				Bdata[1] = Gdata[1] + (Gdata[1] * ColorAbeB[0]);
				Bdata[2] = Gdata[2] + (Gdata[2] * ColorAbeB[0]);
			}
			else // ColorAbeB_Flag < 10000 
			{
				Bdata[0] = Gdata[0] - (Gdata[0] * ColorAbeB[0]);
				Bdata[1] = Gdata[1] - (Gdata[1] * ColorAbeB[0]);
				Bdata[2] = Gdata[2] - (Gdata[2] * ColorAbeB[0]);
			}
			Bdata[3] = 0.000000;


			dng_vector radialParam[3];
			dng_vector tangentialParam[3];
			radialParam[0].fCount = 4;					
			radialParam[1].fCount = 4;	
			radialParam[2].fCount = 4;

			 //GREEN
			radialParam[1].fData[0] = Gdata[0];
			radialParam[1].fData[1] = Gdata[1];
			radialParam[1].fData[2] = Gdata[2];
			radialParam[1].fData[3] = 0.000000;
			
			//RED
			radialParam[0].fData[0] = Rdata[0]; 	
			radialParam[0].fData[1] = Rdata[1];	
			radialParam[0].fData[2] = Rdata[2]; 	
			radialParam[0].fData[3] = 0.000000; 

			//BLUE
			radialParam[2].fData[0] = Bdata[0]; 	
			radialParam[2].fData[1] = Bdata[1]; 	
			radialParam[2].fData[2] = Bdata[2];		
			radialParam[2].fData[3] = 0.000000; 
					
			tangentialParam[0].fCount = 2;		tangentialParam[0].fData[0] = 0.0;		tangentialParam[0].fData[1] = 0.0;
			tangentialParam[1].fCount = 2;		tangentialParam[1].fData[0] = 0.0;		tangentialParam[1].fData[1] = 0.0;
			tangentialParam[2].fCount = 2;		tangentialParam[2].fData[0] = 0.0;		tangentialParam[2].fData[1] = 0.0;

			dng_point_real64 opCenter;
			opCenter.h = 0.5;
			opCenter.v = 0.5;

			dng_warp_params_rectilinear rectParam(3, radialParam, tangentialParam, opCenter);	
			dng_opcode_WarpRectilinear opcodeRect(rectParam, 1);

			AutoPtr<dng_opcode_WarpRectilinear> opcodeRect1;

			dng_opcode *result = new dng_opcode_WarpRectilinear(rectParam, 1);
			AutoPtr<dng_opcode> opcodePtr(result);

			negative->OpcodeList3().Append(opcodePtr);
			//********************************************LDC********************************************//


            // Markernote backup.

            QByteArray mkrnts = meta.getExifTagData("Exif.Photo.MakerNote");

            if (!mkrnts.isEmpty())
            {
                qDebug() << "DNGWriter: Backup Makernote (" << mkrnts.size() << " bytes)" ;

                dng_memory_allocator memalloc(gDefaultDNGMemoryAllocator);
                dng_memory_stream stream(memalloc);
                stream.Put(mkrnts.data(), mkrnts.size());
                AutoPtr<dng_memory_block> block(host.Allocate(mkrnts.size()));
                stream.SetReadPosition(0);
                stream.Get(block->Buffer(), mkrnts.size());

                if (identifyMake.make != "Canon")
                {
                    negative->SetMakerNote(block);
                    negative->SetMakerNoteSafety(true);
                }

                long mknOffset       = 0;
                QString mknByteOrder = meta.getExifTagString("Exif.MakerNote.ByteOrder");

                if ((meta.getExifTagLong("Exif.MakerNote.Offset", mknOffset)) && !mknByteOrder.isEmpty())
                {
                    dng_memory_stream streamPriv(memalloc);
                    streamPriv.SetBigEndian();

                    streamPriv.Put("Adobe", 5);
                    streamPriv.Put_uint8(0x00);
                    streamPriv.Put("MakN", 4);
                    streamPriv.Put_uint32(mkrnts.size() + mknByteOrder.size() + 4);
                    streamPriv.Put(mknByteOrder.toLatin1(), mknByteOrder.size());
                    streamPriv.Put_uint32(mknOffset);
                    streamPriv.Put(mkrnts.data(), mkrnts.size());
                    AutoPtr<dng_memory_block> blockPriv(host.Allocate(streamPriv.Length()));
                    streamPriv.SetReadPosition(0);
                    streamPriv.Get(blockPriv->Buffer(), streamPriv.Length());
                    negative->SetPrivateData(blockPriv);
                }
            }
        }

        if (d->backupOriginalRawFile)
        {
            qDebug() << "DNGWriter: Backup Original RAW file (" << inputInfo.size() << " bytes)";

            QFileInfo originalFileInfo(inputFile());

            QFile originalFile(originalFileInfo.absoluteFilePath());
            originalFile.open(QIODevice::ReadOnly);
            QDataStream originalDataStream(&originalFile);

            quint32 forkLength = originalFileInfo.size();
            quint32 forkBlocks = (quint32)floor((forkLength + 65535.0) / 65536.0);

            QVector<quint32> offsets;
            quint32 offset = (2 + forkBlocks) * sizeof(quint32);
            offsets.push_back(offset);

            QByteArray originalDataBlock;
            originalDataBlock.resize(CHUNK);

            QTemporaryFile compressedFile;

            if (!compressedFile.open())
            {
                qDebug() << "DNGWriter: Cannot open temporary file to write Zip Raw file. Aborted...";
                return PROCESSFAILED;
            }

            QDataStream compressedDataStream(&compressedFile);

            for (quint32 block = 0; block < forkBlocks; ++block)
            {
                int originalBlockLength = originalDataStream.readRawData(originalDataBlock.data(), CHUNK);

                QByteArray compressedDataBlock = qCompress((const uchar*)originalDataBlock.data(), originalBlockLength, -1);
                compressedDataBlock.remove(0, 4); // removes qCompress own header
                qDebug() << "DNGWriter: compressed data block " << originalBlockLength << " -> " << compressedDataBlock.size();

                offset += compressedDataBlock.size();
                offsets.push_back(offset);

                compressedDataStream.writeRawData(compressedDataBlock.data(), compressedDataBlock.size());
            }

            dng_memory_allocator memalloc(gDefaultDNGMemoryAllocator);
            dng_memory_stream tempDataStream(memalloc);
            tempDataStream.SetBigEndian(true);
            tempDataStream.Put_uint32(forkLength);

            for (qint32 idx = 0; idx < offsets.size(); ++idx)
            {
                tempDataStream.Put_uint32(offsets[idx]);
            }

            QByteArray compressedData;
            compressedData.resize(compressedFile.size());
            compressedFile.seek(0);
            compressedDataStream.readRawData(compressedData.data(), compressedData.size());
            tempDataStream.Put(compressedData.data(), compressedData.size());

            compressedFile.remove();

            tempDataStream.Put_uint32(0);
            tempDataStream.Put_uint32(0);
            tempDataStream.Put_uint32(0);
            tempDataStream.Put_uint32(0);
            tempDataStream.Put_uint32(0);
            tempDataStream.Put_uint32(0);
            tempDataStream.Put_uint32(0);

            AutoPtr<dng_memory_block> block(host.Allocate(tempDataStream.Length()));
            tempDataStream.SetReadPosition(0);
            tempDataStream.Get(block->Buffer(), tempDataStream.Length());

            dng_md5_printer md5;
            md5.Process(block->Buffer(), block->LogicalSize());
            negative->SetOriginalRawFileData(block);
            negative->SetOriginalRawFileDigest(md5.Result());
            negative->ValidateOriginalRawFileDigest();
        }

        if (d->cancel)
            return PROCESSCANCELED;
		
		result = (double)(clock() - before) / CLOCKS_PER_SEC;
		qDebug() << " dng setting time " << result ;
		before = clock();
        // -----------------------------------------------------------------------------------------

        qDebug() << "DNGWriter: Build DNG Negative" ;

        // Assign Raw image data.
        negative->SetStage1Image(image);
		
		result = (double)(clock() - before) / CLOCKS_PER_SEC;
		qDebug() << " SetStage1Image time " << result ;
		before = clock();

        // Compute linearized and range mapped image
        negative->BuildStage2Image(host);

		result = (double)(clock() - before) / CLOCKS_PER_SEC;
		qDebug() << " BuildStage2Image time " << result ;
		before = clock();

        // Compute demosaiced image (used by preview and thumbnail)
        negative->BuildStage3Image(host);

		result = (double)(clock() - before) / CLOCKS_PER_SEC;
		qDebug() << " BuildStage3Image time " << result ;
		before = clock();

        negative->SynchronizeMetadata();
        negative->RebuildIPTC(true, false);

		result = (double)(clock() - before) / CLOCKS_PER_SEC;
		qDebug() << " RebuildIPTC time " << result ;
		before = clock();

        if (d->cancel)
            return PROCESSCANCELED;
		qDebug() << "DNGWriter: Build DNG Negative endendend" ;
        // -----------------------------------------------------------------------------------------

        dng_preview_list previewList;

// NOTE: something is wrong with Qt < 4.4.0 to import TIFF data as stream in QImage.
#if QT_VERSION >= 0x40400

        if (d->previewMode != DNGWriter::NONE)
        {
            qDebug() << "DNGWriter: DNG preview image creation" ;

            // Construct a preview image as TIFF format.
            AutoPtr<dng_image> tiffImage;
            dng_render tiff_render(host, *negative);
            tiff_render.SetFinalSpace(dng_space_sRGB::Get());
            tiff_render.SetFinalPixelType(ttByte);
            tiff_render.SetMaximumSize(d->previewMode == MEDIUM ? 1280 : width);
            tiffImage.Reset(tiff_render.Render());

            dng_image_writer tiff_writer;
            AutoPtr<dng_memory_stream> dms(new dng_memory_stream(gDefaultDNGMemoryAllocator));

            tiff_writer.WriteTIFF(host, *dms, *tiffImage.Get(), piRGB,
                                  ccUncompressed, negative.Get(), &tiff_render.FinalSpace());

            // Write TIFF preview image data to a temp JPEG file
            std::vector<char> tiff_mem_buffer(dms->Length());
            dms->SetReadPosition(0);
            dms->Get(&tiff_mem_buffer.front(), tiff_mem_buffer.size());
            dms.Reset();

            QImage pre_image;

			qDebug() << "tiff_mem_buffer.size()" << tiff_mem_buffer.size();

            if (!pre_image.loadFromData((uchar*)&tiff_mem_buffer.front(), tiff_mem_buffer.size(), "TIFF"))
            {
                qDebug() << "DNGWriter: Cannot load TIFF preview data in memory. Aborted..." ;
                return PROCESSFAILED;
            }
			
            QTemporaryFile previewFile;

            if (!previewFile.open())
            {
                qDebug() << "DNGWriter: Cannot open temporary file to write JPEG preview. Aborted..." ;
                return PROCESSFAILED;
            }

            if (!pre_image.save(previewFile.fileName(), "JPEG", 90))
            {
                qDebug() << "DNGWriter: Cannot save file to write JPEG preview. Aborted..." ;
                return PROCESSFAILED;
            }			

            // Load JPEG preview file data in DNG preview container.
            AutoPtr<dng_jpeg_preview> jpeg_preview;
            jpeg_preview.Reset(new dng_jpeg_preview);
            jpeg_preview->fPhotometricInterpretation = piYCbCr;
            jpeg_preview->fPreviewSize.v             = pre_image.height();
            jpeg_preview->fPreviewSize.h             = pre_image.width();
            jpeg_preview->fCompressedData.Reset(host.Allocate(previewFile.size()));

			qDebug() << "jpeg_preview->fPreviewSize.v height  " << jpeg_preview->fPreviewSize.v;
			qDebug() << "jpeg_preview->fPreviewSize.h width  " << jpeg_preview->fPreviewSize.h;
            QDataStream previewStream( &previewFile );
            previewStream.readRawData(jpeg_preview->fCompressedData->Buffer_char(), previewFile.size());

            AutoPtr<dng_preview> pp( dynamic_cast<dng_preview*>(jpeg_preview.Release()) );
            previewList.Append(pp);

            previewFile.remove();
        }

#endif /* QT_VERSION >= 0x40400 */

        if (d->cancel)
            return PROCESSCANCELED;

        // -----------------------------------------------------------------------------------------

        qDebug() << "DNGWriter: DNG thumbnail creation" ;

        dng_image_preview thumbnail;
        dng_render thumbnail_render(host, *negative);
        thumbnail_render.SetFinalSpace(dng_space_sRGB::Get());
        thumbnail_render.SetFinalPixelType(ttByte);
        thumbnail_render.SetMaximumSize(256);
        thumbnail.fImage.Reset(thumbnail_render.Render());

        if (d->cancel)
            return PROCESSCANCELED;

        // -----------------------------------------------------------------------------------------

        qDebug() << "DNGWriter: Creating DNG file " << outputInfo.fileName() ;

        dng_image_writer writer;
        dng_file_stream filestream(QFile::encodeName(dngFilePath), true);		
        
		writer.WriteDNG(host, filestream, *negative.Get(), thumbnail,
                        d->jpegLossLessCompression ? ccJPEG : ccUncompressed,
                        &previewList);

		result = (double)(clock() - before) / CLOCKS_PER_SEC;
		qDebug() << " WriteDNG tiff preview time " << result ;
		before = clock();
        // -----------------------------------------------------------------------------------------
        // Metadata makernote cleanup using Exiv2 for some RAW file types
        // See B.K.O #204437 and #210371, and write XMP Sidecar if necessary

        if (meta.load(dngFilePath))
        {
            if (inputInfo.suffix().toUpper() == QString("ORF"))
            {
                qDebug() << "DNGWriter: cleanup makernotes using Exiv2" ;

                meta.setWriteRawFiles(true);
                meta.removeExifTag("Exif.OlympusIp.BlackLevel", false);
            }
            else
            {
                // Don't touch DNG file and create XMP sidecar depending of KIPI host application settings.
                meta.setWriteRawFiles(false);
            }

            meta.applyChanges();
        }

        // -----------------------------------------------------------------------------------------
        // update modification time if desired

        if (d->updateFileDate)
        {
            QDateTime date = meta.getImageDateTime();

            qDebug() << "DNGWriter: Setting modification date from meta data: " << date.toString();

            // don't touch access time
            struct stat st;
            stat(QFile::encodeName(dngFilePath), &st);
			/*
            struct utimbuf ut;
            ut.actime  = st.st_atime;
            ut.modtime = date.toTime_t();
            utime(QFile::encodeName(dngFilePath), &ut);
			*/
        }
    }

    catch (const dng_exception& exception)
    {
        int ret = exception.ErrorCode();
        qDebug() << "DNGWriter: DNG SDK exception code (" << ret << ")" ;
        return DNGSDKINTERNALERROR;
    }

    catch (...)
    {
        qDebug() << "DNGWriter: DNG SDK exception code unknow" ;
        return DNGSDKINTERNALERROR;
    }

    qDebug() << "DNGWriter: DNG conversion complete..." ;

    return PROCESSCOMPLETE;
}

bool DNGWriter::fujiRotate(QByteArray& rawData, KDcrawIface::DcrawInfoContainer& identify) const
{
    QByteArray tmpData(rawData);
    int height             = identify.outputSize.height();
    int width              = identify.outputSize.width();
    unsigned short* tmp    = reinterpret_cast<unsigned short*>(tmpData.data());
    unsigned short* output = reinterpret_cast<unsigned short*>(rawData.data());

    for (int row=0; row < height; ++row)
    {
        for (int col=0; col < width; ++col)
        {
            output[col * height + row] = tmp[row * width + col];
        }
    }

    identify.orientation = DcrawInfoContainer::ORIENTATION_Mirror90CCW;
    identify.outputSize  = QSize(height, width);
    //TODO: rotate margins
    return true;
}
}  // namespace DNGIface
